# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

import yaml

from swiftclient import exceptions as swiftexceptions
from ironicclient import exceptions as ironicexceptions

from tripleo_common import constants
from tripleo_common.tests import base
from tripleo_common.utils import stack_parameters
from tripleo_common.utils import nodes


class StackParametersTest(base.TestCase):

    @mock.patch('tripleo_common.utils.plan.'
                'cache_delete')
    def test_reset_parameter(self, mock_cache):
        swift = mock.MagicMock(url="http://test.com")
        mock_env = yaml.safe_dump({
            'name': constants.DEFAULT_CONTAINER_NAME,
            'template': 'template',
            'environments': [{u'path': u'environments/test.yaml'}],
            'parameter_defaults': {'SomeTestParameter': 42}
        }, default_flow_style=False)
        swift.get_object.return_value = ({}, mock_env)

        # Test
        stack_parameters.reset_parameters(swift)

        mock_env_reset = yaml.safe_dump({
            'name': constants.DEFAULT_CONTAINER_NAME,
            'template': 'template',
            'environments': [{u'path': u'environments/test.yaml'}]
        }, default_flow_style=False)

        swift.put_object.assert_called_once_with(
            constants.DEFAULT_CONTAINER_NAME,
            constants.PLAN_ENVIRONMENT,
            mock_env_reset
        )
        mock_cache.assert_called_once_with(
            swift,
            "overcloud",
            "tripleo.parameters.get"
        )

    @mock.patch('uuid.uuid4')
    @mock.patch('heatclient.common.template_utils.'
                'process_multiple_environments_and_files')
    @mock.patch('heatclient.common.template_utils.'
                'get_template_contents')
    @mock.patch('tripleo_common.utils.plan.'
                'cache_set')
    def test_update_parameters(self, mock_cache,
                               mock_get_template_contents,
                               mock_env_files,
                               mock_uuid):

        mock_env_files.return_value = ({}, {})

        swift = mock.MagicMock(url="http://test.com")

        mock_env = yaml.safe_dump({
            'name': constants.DEFAULT_CONTAINER_NAME,
            'temp_environment': 'temp_environment',
            'template': 'template',
            'environments': [{u'path': u'environments/test.yaml'}],
        }, default_flow_style=False)

        mock_roles = yaml.safe_dump([{"name": "foo"}])
        mock_network = yaml.safe_dump([{'enabled': False}])
        mock_exclude = yaml.safe_dump({"name": "foo"})

        swift.get_object.side_effect = (
            ({}, mock_env),
            ({}, mock_env),
            ({}, mock_roles),
            ({}, mock_network),
            ({}, mock_exclude),
            ({}, mock_env),
            ({}, mock_env),
            ({}, mock_env),
            ({}, mock_roles),
            ({}, mock_network),
            ({}, mock_exclude),
            ({}, mock_env),
            ({}, mock_env),
            swiftexceptions.ClientException('atest2')
        )

        def return_container_files(*args):
            return ('headers', [{'name': 'foo.role.j2.yaml'}])

        swift.get_container = mock.MagicMock(
            side_effect=return_container_files)

        mock_heat = mock.MagicMock()

        mock_heat.stacks.validate.return_value = {
            "Type": "Foo",
            "Description": "Le foo bar",
            "Parameters": {"bar": {"foo": "bar barz"}},
            "NestedParameters": {"Type": "foobar"}
        }

        mock_uuid.return_value = "cheese"

        expected_value = {
            'environment_parameters': None,
            'heat_resource_tree': {
                'parameters': {'bar': {'foo': 'bar barz',
                                       'name': 'bar'}},
                'resources': {'cheese': {
                    'id': 'cheese',
                    'name': 'Root',
                    'description': 'Le foo bar',
                    'parameters': ['bar'],
                    'resources': ['cheese'],
                    'type': 'Foo'}
                }
            }
        }

        mock_get_template_contents.return_value = ({}, {
            'heat_template_version': '2016-04-30'
        })

        # Test
        test_parameters = {'SomeTestParameter': 42}
        result = stack_parameters.update_parameters(
            swift, mock_heat, test_parameters)

        mock_env_updated = yaml.safe_dump({
            'name': constants.DEFAULT_CONTAINER_NAME,
            'temp_environment': 'temp_environment',
            'parameter_defaults': {'SomeTestParameter': 42},
            'template': 'template',
            'environments': [{u'path': u'environments/test.yaml'}]
        }, default_flow_style=False)

        swift.put_object.assert_any_call(
            constants.DEFAULT_CONTAINER_NAME,
            constants.PLAN_ENVIRONMENT,
            mock_env_updated
        )

        mock_heat.stacks.validate.assert_called_once_with(
            environment={},
            files={},
            show_nested=True,
            template={'heat_template_version': '2016-04-30'},
        )

        mock_cache.assert_called_once_with(
            swift,
            "overcloud",
            "tripleo.parameters.get",
            expected_value
        )
        self.assertEqual(result, expected_value)

    @mock.patch('heatclient.common.template_utils.'
                'process_multiple_environments_and_files')
    @mock.patch('heatclient.common.template_utils.'
                'get_template_contents')
    @mock.patch('tripleo_common.utils.plan.'
                'cache_set')
    def test_update_parameter_new_key(self, mock_cache,
                                      mock_get_template_contents,
                                      mock_env_files):

        mock_env_files.return_value = ({}, {})

        swift = mock.MagicMock(url="http://test.com")
        mock_env = yaml.safe_dump({
            'name': constants.DEFAULT_CONTAINER_NAME,
            'temp_environment': 'temp_environment',
            'template': 'template',
            'environments': [{u'path': u'environments/test.yaml'}],
        }, default_flow_style=False)

        mock_roles = yaml.safe_dump([{"name": "foo"}])
        mock_network = yaml.safe_dump([{'enabled': False}])
        mock_exclude = yaml.safe_dump({"name": "foo"})

        swift.get_object.side_effect = (
            ({}, mock_env),
            ({}, mock_env),
            ({}, mock_roles),
            ({}, mock_network),
            ({}, mock_exclude),
            ({}, mock_env),
            ({}, mock_env),
            ({}, mock_env),
            ({}, mock_roles),
            ({}, mock_network),
            ({}, mock_exclude),
            ({}, mock_env),
            ({}, mock_env),
            swiftexceptions.ClientException('atest2')
        )

        def return_container_files(*args):
            return ('headers', [{'name': 'foo.role.j2.yaml'}])

        swift.get_container = mock.MagicMock(
            side_effect=return_container_files)

        heat = mock.MagicMock()
        heat.stacks.validate.return_value = {}

        mock_get_template_contents.return_value = ({}, {
            'heat_template_version': '2016-04-30'
        })

        # Test
        test_parameters = {'SomeTestParameter': 42}
        stack_parameters.update_parameters(
            swift, heat, test_parameters,
            parameter_key='test_key')
        mock_env_updated = yaml.safe_dump({
            'name': constants.DEFAULT_CONTAINER_NAME,
            'temp_environment': 'temp_environment',
            'test_key': {'SomeTestParameter': 42},
            'template': 'template',
            'environments': [{u'path': u'environments/test.yaml'}]
        }, default_flow_style=False)

        swift.put_object.assert_any_call(
            constants.DEFAULT_CONTAINER_NAME,
            constants.PLAN_ENVIRONMENT,
            mock_env_updated
        )

        heat.stacks.validate.assert_called_once_with(
            environment={},
            files={},
            show_nested=True,
            template={'heat_template_version': '2016-04-30'},
        )

        mock_cache.assert_called_once_with(
            swift,
            "overcloud",
            "tripleo.parameters.get",
            {'environment_parameters': None, 'heat_resource_tree': {}}
        )

    @mock.patch('heatclient.common.template_utils.'
                'process_multiple_environments_and_files')
    @mock.patch('heatclient.common.template_utils.'
                'get_template_contents')
    @mock.patch('tripleo_common.utils.plan.'
                'cache_set')
    @mock.patch('tripleo_common.utils.parameters.set_count_and_flavor_params')
    def test_update_role_parameter(self, mock_set_count_and_flavor,
                                   mock_cache, mock_get_template_contents,
                                   mock_env_files):

        mock_env_files.return_value = ({}, {})

        swift = mock.MagicMock(url="http://test.com")
        mock_env = yaml.safe_dump({
            'name': 'overcast'
        }, default_flow_style=False)

        mock_roles = yaml.safe_dump([{"name": "foo"}])
        mock_network = yaml.safe_dump([{'enabled': False}])
        mock_exclude = yaml.safe_dump({"name": "foo"})

        swift.get_object.side_effect = (
            ({}, mock_env),
            ({}, mock_env),
            ({}, mock_roles),
            ({}, mock_network),
            ({}, mock_exclude),
            ({}, mock_env),
            ({}, mock_env),
            ({}, mock_env),
            ({}, mock_roles),
            ({}, mock_network),
            ({}, mock_exclude),
            ({}, mock_env),
            ({}, mock_env),
            swiftexceptions.ClientException('atest2')
        )

        def return_container_files(*args):
            return ('headers', [{'name': 'foo.yaml'}])

        swift.get_container = mock.MagicMock(
            side_effect=return_container_files)

        heat = mock.MagicMock()
        ironic = mock.MagicMock()
        compute = mock.MagicMock()

        heat.stacks.validate.return_value = {}

        mock_get_template_contents.return_value = ({}, {
            'heat_template_version': '2016-04-30'
        })

        params = {'CephStorageCount': 1,
                  'OvercloudCephStorageFlavor': 'ceph-storage'}
        mock_set_count_and_flavor.return_value = params

        stack_parameters.update_role_parameters(
            swift, heat, ironic, compute,
            'ceph-storage', 'overcast')
        mock_env_updated = yaml.safe_dump({
            'name': 'overcast',
            'parameter_defaults': params
        }, default_flow_style=False)

        swift.put_object.assert_any_call(
            'overcast',
            constants.PLAN_ENVIRONMENT,
            mock_env_updated
        )

        heat.stacks.validate.assert_called_once_with(
            environment={},
            files={},
            show_nested=True,
            template={'heat_template_version': '2016-04-30'},
        )

        mock_cache.assert_called_once_with(
            swift,
            "overcast",
            "tripleo.parameters.get",
            {'environment_parameters': None, 'heat_resource_tree': {}}
        )

    @mock.patch('tripleo_common.utils.plan.'
                'cache_set')
    @mock.patch('tripleo_common.utils.plan.'
                'cache_get')
    @mock.patch('heatclient.common.template_utils.'
                'process_multiple_environments_and_files')
    @mock.patch('heatclient.common.template_utils.get_template_contents')
    def test_empty_resource_tree(self,
                                 mock_get_template_contents,
                                 mock_process_multiple_environments_and_files,
                                 mock_cache_get,
                                 mock_cache_set):

        mock_cache_get.return_value = None
        swift = mock.MagicMock(url="http://test.com")
        mock_env = yaml.safe_dump({
            'temp_environment': 'temp_environment',
            'template': 'template',
            'environments': [{u'path': u'environments/test.yaml'}]
        }, default_flow_style=False)
        swift.get_object.side_effect = (
            ({}, mock_env),
            swiftexceptions.ClientException('atest2'),
            ({}, mock_env)
        )
        mock_get_template_contents.return_value = ({}, {
            'heat_template_version': '2016-04-30'
        })
        mock_process_multiple_environments_and_files.return_value = ({}, {})

        mock_heat = mock.MagicMock()
        mock_heat.stacks.validate.return_value = {}

        expected_value = {
            'heat_resource_tree': {},
            'environment_parameters': None,
        }

        # Test
        result = stack_parameters.get_flattened_parameters(swift, mock_heat)
        mock_heat.stacks.validate.assert_called_once_with(
            environment={},
            files={},
            show_nested=True,
            template={'heat_template_version': '2016-04-30'},
        )
        self.assertEqual(result, expected_value)

    @mock.patch('tripleo_common.utils.plan.'
                'cache_set')
    @mock.patch('tripleo_common.utils.plan.'
                'cache_get')
    @mock.patch('uuid.uuid4', side_effect=['1', '2'])
    @mock.patch('heatclient.common.template_utils.'
                'process_multiple_environments_and_files')
    @mock.patch('heatclient.common.template_utils.get_template_contents')
    def test_valid_resource_tree(self,
                                 mock_get_template_contents,
                                 mock_process_multiple_environments_and_files,
                                 mock_uuid,
                                 mock_cache_get,
                                 mock_cache_set):

        mock_cache_get.return_value = None
        swift = mock.MagicMock(url="http://test.com")
        mock_env = yaml.safe_dump({
            'temp_environment': 'temp_environment',
            'template': 'template',
            'environments': [{u'path': u'environments/test.yaml'}]
        }, default_flow_style=False)
        swift.get_object.side_effect = (
            ({}, mock_env),
            swiftexceptions.ClientException('atest2'),
            ({}, mock_env)
        )
        mock_get_template_contents.return_value = ({}, {
            'heat_template_version': '2016-04-30'
        })
        mock_process_multiple_environments_and_files.return_value = ({}, {})

        mock_heat = mock.MagicMock()
        mock_heat.stacks.validate.return_value = {
            'NestedParameters': {
                'CephStorageHostsDeployment': {
                    'Type': 'OS::Heat::StructuredDeployments',
                },
            },
            'description': 'sample',
            'Parameters': {
                'ControllerCount': {
                    'Default': 1,
                    'Type': 'Number',
                },
            }
        }

        expected_value = {
            'heat_resource_tree': {
                'resources': {
                    '1': {
                        'id': '1',
                        'name': 'Root',
                        'resources': [
                            '2'
                        ],
                        'parameters': [
                            'ControllerCount'
                        ]
                    },
                    '2': {
                        'id': '2',
                        'name': 'CephStorageHostsDeployment',
                        'type': 'OS::Heat::StructuredDeployments'
                    }
                },
                'parameters': {
                    'ControllerCount': {
                        'default': 1,
                        'type': 'Number',
                        'name': 'ControllerCount'
                    }
                },
            },
            'environment_parameters': None,
        }

        # Test
        result = stack_parameters.get_flattened_parameters(swift, mock_heat)
        self.assertEqual(result, expected_value)

    def test_generate_hostmap(self):

        # two instances in 'nova list'.
        # vm1 with id=123 and vm2 with id=234
        server1 = mock.MagicMock()
        server1.id = 123
        server1.name = 'vm1'

        server2 = mock.MagicMock()
        server2.id = 234
        server2.name = 'vm2'

        servers = mock.MagicMock()
        servers = [server1, server2]

        compute_client = mock.MagicMock()
        compute_client.servers.list.side_effect = (servers, )

        # we assume instance id=123 has been provisioned using bm node 'bm1'
        # while instance id=234 is in error state, so no bm node has been used

        def side_effect(args):
            if args == 123:
                return bm1
            elif args == 234:
                raise ironicexceptions.NotFound

        baremetal_client = mock.MagicMock()
        baremetal_client.node.get_by_instance_uuid = mock.MagicMock(
            side_effect=side_effect)

        # bm server with name='bm1' and uuid='9876'
        bm1 = mock.MagicMock()
        bm1.uuid = 9876
        bm1.name = 'bm1'

        # 'bm1' has a single port with mac='aa:bb:cc:dd:ee:ff'
        port1 = mock.MagicMock()
        port1.address = 'aa:bb:cc:dd:ee:ff'

        def side_effect2(node, *args):
            if node == 9876:
                return [port1, ]
            else:
                raise ironicexceptions.NotFound

        baremetal_client.port.list = mock.MagicMock(side_effect=side_effect2)

        expected_hostmap = {
            'aa:bb:cc:dd:ee:ff': {
                'compute_name': 'vm1',
                'baremetal_name': 'bm1'
                }
            }

        result = nodes.generate_hostmap(baremetal_client, compute_client)
        self.assertEqual(result, expected_hostmap)

    @mock.patch('tripleo_common.utils.nodes.generate_hostmap')
    def test_generate_fencing_parameters(self, mock_generate_hostmap):
        test_hostmap = {
            "00:11:22:33:44:55": {
                "compute_name": "compute_name_0",
                "baremetal_name": "baremetal_name_0"
                },
            "11:22:33:44:55:66": {
                "compute_name": "compute_name_1",
                "baremetal_name": "baremetal_name_1"
                },
            "aa:bb:cc:dd:ee:ff": {
                "compute_name": "compute_name_4",
                "baremetal_name": "baremetal_name_4"
                },
            "bb:cc:dd:ee:ff:gg": {
                "compute_name": "compute_name_5",
                "baremetal_name": "baremetal_name_5"
                }
            }
        mock_generate_hostmap.return_value = test_hostmap

        test_envjson = [{
            "name": "control-0",
            "pm_password": "control-0-password",
            "pm_type": "ipmi",
            "pm_user": "control-0-admin",
            "pm_addr": "0.1.2.3",
            "pm_port": "0123",
            "mac": [
                "00:11:22:33:44:55"
            ]
        }, {
            "name": "control-1",
            "pm_password": "control-1-password",
            # Still support deprecated drivers
            "pm_type": "pxe_ipmitool",
            "pm_user": "control-1-admin",
            "pm_addr": "1.2.3.4",
            "mac": [
                "11:22:33:44:55:66"
            ]
        }, {
            # test node using redfish pm
            "name": "compute-4",
            "pm_password": "calvin",
            "pm_type": "redfish",
            "pm_user": "root",
            "pm_addr": "172.16.0.1:8000",
            "pm_port": "8000",
            "redfish_verify_ca": "false",
            "pm_system_id": "/redfish/v1/Systems/5678",
            "mac": [
                "aa:bb:cc:dd:ee:ff"
            ]
        }, {
            # This is an extra node on oVirt/RHV
            "name": "control-3",
            "pm_password": "ovirt-password",
            "pm_type": "staging-ovirt",
            "pm_user": "admin@internal",
            "pm_addr": "3.4.5.6",
            "pm_vm_name": "control-3",
            "mac": [
                "bb:cc:dd:ee:ff:gg"
            ]
        }, {
            # This is an extra node that is not in the hostmap, to ensure we
            # cope with unprovisioned nodes
            "name": "control-2",
            "pm_password": "control-2-password",
            "pm_type": "ipmi",
            "pm_user": "control-2-admin",
            "pm_addr": "2.3.4.5",
            "mac": [
                "22:33:44:55:66:77"
            ]
        }
        ]

        ironic = mock.MagicMock()
        compute = mock.MagicMock()

        result = stack_parameters.generate_fencing_parameters(
            ironic, compute, test_envjson,
            28, 5, 0, True)['parameter_defaults']

        self.assertTrue(result["EnableFencing"])
        self.assertEqual(len(result["FencingConfig"]["devices"]), 4)
        self.assertEqual(result["FencingConfig"]["devices"][0], {
                         "agent": "fence_ipmilan",
                         "host_mac": "00:11:22:33:44:55",
                         "params": {
                             "delay": 28,
                             "ipaddr": "0.1.2.3",
                             "ipport": "0123",
                             "lanplus": True,
                             "privlvl": 5,
                             "login": "control-0-admin",
                             "passwd": "control-0-password",
                             "pcmk_host_list": "compute_name_0"
                             }
                         })
        self.assertEqual(result["FencingConfig"]["devices"][1], {
                         "agent": "fence_ipmilan",
                         "host_mac": "11:22:33:44:55:66",
                         "params": {
                             "delay": 28,
                             "ipaddr": "1.2.3.4",
                             "lanplus": True,
                             "privlvl": 5,
                             "login": "control-1-admin",
                             "passwd": "control-1-password",
                             "pcmk_host_list": "compute_name_1"
                             }
                         })
        self.assertEqual(result["FencingConfig"]["devices"][2], {
                         "agent": "fence_redfish",
                         "host_mac": "aa:bb:cc:dd:ee:ff",
                         "params": {
                             "delay": 28,
                             "ipaddr": "172.16.0.1:8000",
                             "ipport": "8000",
                             "privlvl": 5,
                             "login": "root",
                             "passwd": "calvin",
                             "systems_uri": "/redfish/v1/Systems/5678",
                             "ssl_insecure": "true",
                             "pcmk_host_list": "compute_name_4"
                             }
                         })
        self.assertEqual(result["FencingConfig"]["devices"][3], {
                         "agent": "fence_rhevm",
                         "host_mac": "bb:cc:dd:ee:ff:gg",
                         "params": {
                             "delay": 28,
                             "ipaddr": "3.4.5.6",
                             "login": "admin@internal",
                             "passwd": "ovirt-password",
                             "port": "control-3",
                             "ssl": 1,
                             "ssl_insecure": 1,
                             "pcmk_host_list": "compute_name_5"
                             }
                         })

    @mock.patch('tripleo_common.utils.template.process_templates')
    def test_run_valid_network_config(self, mock_process_templates):
        swift = mock.MagicMock(url="http://test.com")

        mock_env = {
            'template': {},
            'files': {},
            'environment': [{'path': 'environments/test.yaml'}]
        }

        mock_process_templates.return_value = mock_env
        mock_heat = mock.MagicMock()

        mock_heat.stacks.preview.return_value = mock.Mock(resources=[{
            "resource_identity": {"stack_name": "overcloud-TEMP-Compute-0"},
            "resource_name": "OsNetConfigImpl",
            "properties": {"config": "echo \'{\"network_config\": {}}\'"
                           " > /etc/os-net-config/config.json"}
            }])

        mock_heat.stacks.get.return_value = mock.Mock(
            stack_name="overcloud-TEMP")

        expected = {"network_config": {}}
        # Test
        result = stack_parameters.get_network_configs(
            swift, mock_heat, container='overcloud', role_name='Compute')
        self.assertEqual(expected, result)
        mock_heat.stacks.preview.assert_called_once_with(
            environment=[{'path': 'environments/test.yaml'}],
            files={},
            template={},
            stack_name='overcloud-TEMP',
        )

    @mock.patch('tripleo_common.utils.template.process_templates')
    def test_run_invalid_network_config(self, mock_process_templates):
        swift = mock.MagicMock(url="http://test.com")

        mock_env = {
            'template': {},
            'files': {},
            'environment': [{'path': 'environments/test.yaml'}]
        }

        mock_process_templates.return_value = mock_env
        mock_heat = mock.MagicMock()

        mock_heat.stacks.preview.return_value = mock.Mock(resources=[{
            "resource_identity": {"stack_name": "overcloud-TEMP-Compute-0"},
            "resource_name": "OsNetConfigImpl",
            "properties": {"config": ""}
            }])

        mock_heat.stacks.get.return_value = mock.Mock(
            stack_name="overcloud-TEMP")

        # Test
        self.assertRaises(RuntimeError,
                          stack_parameters.get_network_configs, swift,
                          mock_heat, container='overcloud',
                          role_name='Compute')
        mock_heat.stacks.preview.assert_called_once_with(
            environment=[{'path': 'environments/test.yaml'}],
            files={},
            template={},
            stack_name='overcloud-TEMP',
        )

    @mock.patch('tripleo_common.utils.template.process_templates')
    def test_run_valid_network_config_with_no_if_routes_inputs(
        self, mock_process_templates):
        swift = mock.MagicMock(url="http://test.com")

        mock_env = {
            'template': {
                'resources': {
                    'ComputeGroupVars': {
                        'properties': {
                            'value': {
                                'role_networks': ['InternalApi',
                                                  'Storage']}
                        }
                    }
                }
            },
            'files': {},
            'environment': {'parameter_defaults': {}}
        }

        mock_process_templates.return_value = mock_env
        mock_heat = mock.MagicMock()

        mock_heat.stacks.preview.return_value = mock.Mock(resources=[{
            "resource_identity": {"stack_name": "overcloud-TEMP-Compute-0"},
            "resource_name": "OsNetConfigImpl",
            "properties": {"config": "echo \'{\"network_config\": {}}\'"
                           " > /etc/os-net-config/config.json"}
            }])

        mock_heat.stacks.get.return_value = mock.Mock(
            stack_name="overcloud-TEMP")

        expected = {"network_config": {}}
        # Test
        result = stack_parameters.get_network_configs(
            swift, mock_heat, container='overcloud', role_name='Compute')
        self.assertEqual(expected, result)
        mock_heat.stacks.preview.assert_called_once_with(
            environment={
                'parameter_defaults': {
                    'InternalApiInterfaceRoutes': [[]],
                    'StorageInterfaceRoutes': [[]]
                }
            },
            files={},
            template={'resources': {'ComputeGroupVars': {'properties': {
                'value': {'role_networks': ['InternalApi', 'Storage']}
                }}}},
            stack_name='overcloud-TEMP',
        )

    @mock.patch('tripleo_common.utils.template.process_templates')
    def test_run_valid_network_config_with_if_routes_inputs(
        self, mock_process_templates):
        swift = mock.MagicMock(url="http://test.com")

        mock_env = {
            'template': {
                'resources': {
                    'ComputeGroupVars': {
                        'properties': {
                            'value': {
                                'role_networks': ['InternalApi',
                                                  'Storage']}
                        }
                    }
                }
            },
            'files': {},
            'environment': {
                'parameter_defaults': {
                    'InternalApiInterfaceRoutes': ['test1'],
                    'StorageInterfaceRoutes': ['test2']
                }}
        }

        mock_process_templates.return_value = mock_env
        mock_heat = mock.MagicMock()

        mock_heat.stacks.preview.return_value = mock.Mock(resources=[{
            "resource_identity": {"stack_name": "overcloud-TEMP-Compute-0"},
            "resource_name": "OsNetConfigImpl",
            "properties": {"config": "echo \'{\"network_config\": {}}\'"
                           " > /etc/os-net-config/config.json"}
            }])

        mock_heat.stacks.get.return_value = mock.Mock(
            stack_name="overcloud-TEMP")

        expected = {"network_config": {}}
        # Test
        result = stack_parameters.get_network_configs(
            swift, mock_heat, container='overcloud', role_name='Compute')
        self.assertEqual(expected, result)
        mock_heat.stacks.preview.assert_called_once_with(
            environment={
                'parameter_defaults': {
                    'InternalApiInterfaceRoutes': ['test1'],
                    'StorageInterfaceRoutes': ['test2']
                }
            },
            files={},
            template={'resources': {'ComputeGroupVars': {'properties': {
                'value': {'role_networks': ['InternalApi', 'Storage']}
                }}}},
            stack_name='overcloud-TEMP',
        )



# Copyright 2016 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import collections
import logging

from mistral_lib import actions

from tripleo_common.actions import base
from tripleo_common.actions import parameters as parameters_actions
from tripleo_common import constants
from tripleo_common import update
from tripleo_common.utils import template as template_utils

LOG = logging.getLogger(__name__)


def get_group_resources_after_delete(groupname, res_to_delete, resources):
    group = next(res for res in resources if
                 res.resource_name == groupname and
                 res.resource_type == constants.RESOURCE_GROUP_TYPE)
    members = []
    for res in resources:
        stack_name, stack_id = next(
            x['href'] for x in res.links if
            x['rel'] == 'stack').rsplit('/', 2)[1:]
        # desired new count of nodes after delete operation should be
        # count of all existing nodes in ResourceGroup which are not
        # in set of nodes being deleted. Also nodes in any delete state
        # from a previous failed update operation are not included in
        # overall count (if such nodes exist)
        if (stack_id == group.physical_resource_id and
            res not in res_to_delete and
                not res.resource_status.startswith('DELETE')):

            members.append(res)

    return members


class ScaleDownAction(base.TripleOAction):
    """Deletes overcloud nodes

    Before calling this method, ensure you have updated the plan
    with any templates or environment files as needed.
    """

    def __init__(self, timeout, nodes=[],
                 container=constants.DEFAULT_CONTAINER_NAME):
        self.container = container
        self.nodes = nodes
        self.timeout_mins = timeout
        super(ScaleDownAction, self).__init__()

    def _update_stack(self, parameters={},
                      timeout_mins=constants.STACK_TIMEOUT_DEFAULT,
                      context=None):
        heat = self.get_orchestration_client(context)
        swift = self.get_object_client(context)
        # TODO(rbrady): migrate _update_stack to it's own action and update
        # the workflow for scale down

        # update the plan parameters with the scaled down parameters
        update_params_action = parameters_actions.UpdateParametersAction(
            parameters, self.container)
        updated_plan = update_params_action.run(context)
        if isinstance(updated_plan, actions.Result):
            return updated_plan

        processed_data = template_utils.process_templates(
            swift, heat, container=self.container
        )
        update.add_breakpoints_cleanup_into_env(processed_data['environment'])

        fields = processed_data.copy()
        fields['timeout_mins'] = timeout_mins
        fields['existing'] = True
        # As we do a PATCH update when deleting nodes, parameters set for a
        # stack before upgrade to newton (ex. ComputeRemovalPolicies),
        # would still take precedence over the ones set in parameter_defaults
        # after upgrade. Clear these parameters for backward compatibility.
        fields['clear_parameters'] = list(parameters.keys())

        LOG.debug('stack update params: %s', fields)
        heat.stacks.update(self.container, **fields)

    def _get_removal_params_from_heat(self, resources_by_role, resources):
        stack_params = {}
        for role, role_resources in resources_by_role.items():
            param_name = "{0}Count".format(role)

            # get real count of nodes for each role. *Count stack parameters
            # can not be used because stack parameters return parameters
            # passed by user no matter if previous update operation succeeded
            # or not
            group_members = get_group_resources_after_delete(
                role, role_resources, resources)
            stack_params[param_name] = str(len(group_members))

            # add instance resource names into removal_policies
            # so heat knows which instances should be removed
            removal_param = "{0}RemovalPolicies".format(role)
            stack_params[removal_param] = [{
                'resource_list': [r.resource_name for r in role_resources]
            }]

            # force reset the removal_policies_mode to 'append'
            # as 'update' can lead to deletion of unintended nodes.
            removal_mode = "{0}RemovalPoliciesMode".format(role)
            stack_params[removal_mode] = 'append'

        return stack_params

    def _match_hostname(self, heatclient, instance_list, res, stack_name):
        type_patterns = ['DeployedServer', 'Server']
        if any(res.resource_type.endswith(x) for x in type_patterns):
            res_details = heatclient.resources.get(
                stack_name, res.resource_name)
            if 'name' in res_details.attributes:
                try:
                    instance_list.remove(res_details.attributes['name'])
                    return True
                except ValueError:
                    return False
        return False

    def run(self, context):
        heatclient = self.get_orchestration_client(context)
        resources = heatclient.resources.list(self.container, nested_depth=5)
        resources_by_role = collections.defaultdict(list)
        instance_list = list(self.nodes)

        for res in resources:
            stack_name, stack_id = next(
                x['href'] for x in res.links if
                x['rel'] == 'stack').rsplit('/', 2)[1:]

            try:
                instance_list.remove(res.physical_resource_id)
            except ValueError:
                if not self._match_hostname(
                    heatclient, instance_list, res, stack_name):
                    continue

            # get resource to remove from resource group (it's parent resource
            # of nova server)
            role_resource = next(x for x in resources if
                                 x.physical_resource_id == stack_id)
            # get the role name which is parent resource name in Heat
            role = role_resource.parent_resource
            resources_by_role[role].append(role_resource)

        resources_by_role = dict(resources_by_role)

        if instance_list:
            raise ValueError(
                "Couldn't find following instances in stack %s: %s" %
                (self.container, ','.join(instance_list)))

        # decrease count for each role (or resource group) and set removal
        # policy for each resource group
        stack_params = self._get_removal_params_from_heat(
            resources_by_role, resources)

        return self._update_stack(parameters=stack_params, context=context)



#   Copyright 2019 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#


from concurrent import futures
import os
import pathlib
import six
import tenacity

from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_log import log as logging

from tripleo_common import constants
from tripleo_common.image.builder import base
from tripleo_common.utils import process

CONF = cfg.CONF
LOG = logging.getLogger(__name__ + ".BuildahBuilder")


class BuildahBuilder(base.BaseBuilder):
    """Builder to build container images with Buildah."""

    log = LOG

    def __init__(self, work_dir, deps, base='fedora', img_type='binary',
                 tag='latest', namespace='master',
                 registry_address='127.0.0.1:8787', push_containers=True,
                 volumes=[], excludes=[], build_timeout=None, debug=False):
        """Setup the parameters to build with Buildah.

        :params work_dir: Directory where the Dockerfiles or Containerfiles
            are generated by Kolla.
        :params deps: Dictionary defining the container images
            dependencies.
        :params base: Base image on which the containers are built.
            Default to fedora.
        :params img_type: Method used to build the image. All TripleO images
            are built from binary method. Can be set to false to remove it
            from image name.
        :params tag: Tag used to identify the images that we build.
            Default to latest.
        :params namespace: Namespace used to build the containers.
            Default to master.
        :params registry_address: IP + port of the registry where we push
            the images. Default is 127.0.0.1:8787.
        :params push: Flag to bypass registry push if False. Default is True
        :params volumes: Bind mount volumes used during buildah bud.
            Default to [].
        :params excludes: List of images to skip. Default to [].
        :params build_timeout: Timeout. Default to constants.BUILD_TIMEOUT
        :params debug: Enable debug flag. Default to False.
        """

        logging.register_options(CONF)
        if debug:
            CONF.debug = True
        logging.setup(CONF, '')

        super(BuildahBuilder, self).__init__()
        if build_timeout is None:
            self.build_timeout = constants.BUILD_TIMEOUT
        else:
            self.build_timeout = build_timeout
        self.work_dir = work_dir
        self.deps = deps
        self.base = base
        self.img_type = img_type
        self.tag = tag
        self.namespace = namespace
        self.registry_address = registry_address
        self.push_containers = push_containers
        self.volumes = volumes
        self.excludes = excludes
        self.debug = debug
        # Each container image has a Dockerfile or a Containerfile.
        # Buildah needs to know the base directory later.
        self.cont_map = {os.path.basename(root): root for root, dirs,
                         fnames in os.walk(self.work_dir)
                         if 'Dockerfile' in fnames or
                         'Containerfile' in fnames}
        # Building images with root so overlayfs is used, and not fuse-overlay
        # from userspace, which would be slower.
        self.buildah_cmd = ['sudo', 'buildah']
        if self.debug:
            self.buildah_cmd.append('--log-level debug')

    def _find_container_dir(self, container_name):
        """Return the path of the Dockerfile/Containerfile directory.

        :params container_name: Name of the container.
        """

        if container_name not in self.cont_map:
            self.log.error('Container not found in Kolla '
                           'deps: %s' % container_name)
        return self.cont_map.get(container_name, '')

    def _get_destination(self, container_name):
        """Return the destination of a container image to push.

        :params container_name: Name of the container.
        """

        destination = "{}/{}/{}".format(
            self.registry_address,
            self.namespace,
            self.base,
        )
        if self.img_type:
            destination += '-' + self.img_type
        destination += '-' + container_name + ':' + self.tag
        return destination

    def _generate_container(self, container_name):
        """Generate a container image by building and pushing the image.

        :params container_name: Name of the container.
        """

        if container_name in self.excludes:
            return

        # NOTE(mwhahaha): Use a try catch block so we can better log issues
        # as this is called in a multiprocess fashion so the exception
        # loses some information when it reaches _multi_build
        try:
            self.build(container_name,
                       self._find_container_dir(container_name))
            if self.push_containers:
                self.push(self._get_destination(container_name))
        except Exception as e:
            self.log.exception(e)
            raise

    @tenacity.retry(
        # Retry up to 5 times: 0, 1, 5, 21, 85
        # http://exponentialbackoffcalculator.com/
        reraise=True,
        wait=tenacity.wait_random_exponential(multiplier=4, max=60),
        stop=tenacity.stop_after_attempt(5),
        before_sleep=tenacity.after_log(LOG, logging.WARNING)
    )
    def build(self, container_name, container_build_path):
        """Build an image from a given directory.

        :params container_name: Name of the container.
        :params container_build_path: Directory where the Dockerfile or
            Containerfile and other files are located to build the image.
        """

        # 'buildah bud' is the command we want because Kolla uses Dockefile to
        # build images.
        # TODO(emilien): Stop ignoring TLS. The deployer should either secure
        # the registry or add it to insecure_registries.
        logfile = container_build_path + '/' + container_name + '-build.log'

        # TODO(ramishra) Hack to make the logfile readable by current user,
        # as we're running buildah as root. This would be removed once we
        # move to rootless buildah.
        pathlib.Path(logfile).touch()

        bud_args = ['bud']
        for v in self.volumes:
            bud_args.extend(['--volume', v])
        # TODO(aschultz): drop --format docker when oci format is properly
        # supported by the undercloud registry
        bud_args.extend(['--format', 'docker', '--tls-verify=False',
                         '--logfile', logfile, '-t',
                         self._get_destination(container_name),
                         container_build_path])
        args = self.buildah_cmd + bud_args
        self.log.info("Building %s image with: %s" %
                      (container_name, ' '.join(args)))
        process.execute(
            *args,
            check_exit_code=True,
            run_as_root=False,
            use_standard_locale=True
        )

    @tenacity.retry(  # Retry up to 10 times with jittered exponential backoff
        reraise=True,
        wait=tenacity.wait_random_exponential(multiplier=1, max=15),
        stop=tenacity.stop_after_attempt(10),
        before_sleep=tenacity.after_log(LOG, logging.WARNING)
    )
    def push(self, destination):
        """Push an image to a container registry.

        :params destination: URL to used to push the container. It contains
            the registry address, namespace, base, img_type (optional),
            container name and tag.
        """
        # TODO(emilien): Stop ignoring TLS. The deployer should either secure
        # the registry or add it to insecure_registries.
        # TODO(emilien) We need to figure out how we can push to something
        # else than a Docker registry.
        args = self.buildah_cmd + ['push', '--tls-verify=False', destination,
                                   'docker://' + destination]
        self.log.info("Pushing %s image with: %s" %
                      (destination, ' '.join(args)))
        if self.debug:
            # buildah push logs to stderr, since there is no --log* opt
            # so we'll use the current logging context for that
            process.execute(*args, log_stdout=True, run_as_root=False,
                            use_standard_locale=True, logger=self.log,
                            loglevel=logging.DEBUG)
        else:
            process.execute(*args, run_as_root=False,
                            use_standard_locale=True)

    def build_all(self, deps=None):
        """Build all containers.

        This function will thread the build process allowing it to complete
        in the shortest possible time.

        :params deps: Dictionary defining the container images
                      dependencies.
        """

        if deps is None:
            deps = self.deps

        container_deps = self._generate_deps(deps=deps, containers=list())
        self.log.debug("All container deps: {}".format(container_deps))
        for containers in container_deps:
            self.log.info("Processing containers: {}".format(containers))
            if isinstance(deps, (list,)):
                self._multi_build(containers=containers)
            else:
                self._multi_build(containers=[containers])

    def _generate_deps(self, deps, containers, prio_list=None):
        """Browse containers dependencies and return an an array.

        When the dependencies are generated they're captured in an array,
        which contains additional arrays. This data structure is later
        used in a futures queue.

        :params deps: Dictionary defining the container images
                      dependencies.
        :params containers: List used to keep track of dependent containers.
        :params prio_list: List used to keep track of nested dependencies.
        :returns: list
        """

        self.log.debug("Process deps: {}".format(deps))
        if isinstance(deps, (six.string_types,)):
            if prio_list:
                prio_list.append(deps)
            else:
                containers.append([deps])

        elif isinstance(deps, (dict,)):
            parents = list(deps.keys())
            if prio_list:
                prio_list.extend(parents)
            else:
                containers.append(parents)
            for value in deps.values():
                self.log.debug("Recursing with: {}".format(value))
                self._generate_deps(
                    deps=value,
                    containers=containers
                )

        elif isinstance(deps, (list,)):
            dep_list = list()
            dep_rehash_list = list()
            for item in deps:
                if isinstance(item, (six.string_types,)):
                    dep_list.append(item)
                else:
                    dep_rehash_list.append(item)

            if dep_list:
                containers.append(dep_list)

            for item in dep_rehash_list:
                self.log.debug("Recursing with: {}".format(item))
                self._generate_deps(
                    deps=item,
                    containers=containers,
                    prio_list=dep_list
                )

        self.log.debug("Constructed containers: {}".format(containers))
        return containers

    def _multi_build(self, containers):
        """Build mutliple containers.

        Multi-thread the build process for all containers defined within
        the containers list.

        :params containers: List defining the container images.
        """

        # Workers will use the processor core count with a max of 8. If
        # the containers array has a length less-than the expected processor
        # count, the workers will be adjusted to meet the expectations of the
        # work being processed.
        workers = min(
            min(
                8,
                max(
                    2,
                    processutils.get_worker_count()
                )
            ),
            len(containers)
        )
        with futures.ThreadPoolExecutor(max_workers=workers) as executor:
            future_to_build = {
                executor.submit(
                    self._generate_container, container_name
                ): container_name for container_name in containers
            }
            done, not_done = futures.wait(
                future_to_build,
                timeout=self.build_timeout,
                return_when=futures.FIRST_EXCEPTION
            )

        # NOTE(cloudnull): Once the job has been completed all completed
        #                  jobs are checked for exceptions. If any jobs
        #                  failed a SystemError will be raised using the
        #                  exception information. If any job was loaded
        #                  but not executed a SystemError will be raised.
        exceptions = list()
        for job in done:
            if job._exception:
                exceptions.append(
                    "\nException information: {exception}".format(
                        exception=job._exception
                    )
                )
        else:
            if exceptions:
                raise RuntimeError(
                    '\nThe following errors were detected during '
                    'container build(s):\n{exceptions}'.format(
                        exceptions='\n'.join(exceptions)
                    )
                )

            if not_done:
                error_msg = (
                    'The following jobs were incomplete: {}'.format(
                        [future_to_build[job] for job in not_done]
                    )
                )

                jobs_with_exceptions = [{
                    'container': future_to_build[job],
                    'exception': job._exception}
                    for job in not_done if job._exception]
                if jobs_with_exceptions:
                    for job_with_exception in jobs_with_exceptions:
                        error_msg = error_msg + os.linesep + (
                            "%(container)s raised the following "
                            "exception: %(exception)s" %
                            job_with_exception)

                raise SystemError(error_msg)



#   Copyright 2015 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest import mock

from tripleo_common.image.base import BaseImageManager
from tripleo_common.image.exception import ImageSpecificationException
from tripleo_common.tests import base as testbase
from tripleo_common.tests.image import fakes


class TestBaseImageManager(testbase.TestCase):
    def setUp(self):
        super(TestBaseImageManager, self).setUp()

    @mock.patch('yaml.safe_load', autospec=True)
    @mock.patch('os.path.isfile', autospec=True)
    def test_load_config_files(self, mock_os_path_isfile, mock_yaml_load):
        mock_yaml_load.return_value = fakes.create_disk_images()

        mock_os_path_isfile.return_value = True

        mock_open_context = mock.mock_open()
        mock_open_context().read.return_value = "YAML"

        with mock.patch('six.moves.builtins.open', mock_open_context):
            base_manager = BaseImageManager(['yamlfile'])
            disk_images = base_manager.load_config_files('disk_images')

        mock_yaml_load.assert_called_once_with("YAML")
        self.assertEqual([{
            'arch': 'amd64',
            'distro': 'some_awesome_os',
            'imagename': 'overcloud',
            'type': 'qcow2',
            'elements': ['image_element']
        }], disk_images)

    def test_load_config_files_not_found(self):
        base_manager = BaseImageManager(['file/does/not/exist'])
        self.assertRaises(IOError, base_manager.load_config_files,
                          'disk_images')

    @mock.patch('tripleo_common.image.base.BaseImageManager.APPEND_ATTRIBUTES',
                ['elements', 'options', 'packages', 'environment'])
    @mock.patch('yaml.safe_load', autospec=True)
    @mock.patch('os.path.isfile', autospec=True)
    def test_load_config_files_multiple_files(self, mock_os_path_isfile,
                                              mock_yaml_load):
        mock_yaml_load.side_effect = [{
            'disk_images': [{
                'arch': 'amd64',
                'imagename': 'overcloud',
                'distro': 'some_awesome_distro',
                'type': 'qcow2',
                'elements': ['image_element'],
                'environment': {'test_env': '1'},
            }]},
            {
            'disk_images': [{
                'imagename': 'overcloud',
                'elements': ['another_image_element'],
                'packages': ['a_package'],
                'otherkey': 'some_other_key',
                'environment': {'test_env2': '0'},
            }]}]

        mock_os_path_isfile.return_value = True

        mock_open_context = mock.mock_open()
        mock_open_context().read.return_value = "YAML"

        with mock.patch('six.moves.builtins.open', mock_open_context):
            base_manager = BaseImageManager(['yamlfile1', 'yamlfile2'])
            disk_images = base_manager.load_config_files('disk_images')

        self.assertEqual(2, mock_yaml_load.call_count)
        self.assertEqual([{
            'arch': 'amd64',
            'distro': 'some_awesome_distro',
            'imagename': 'overcloud',
            'type': 'qcow2',
            'elements': ['image_element', 'another_image_element'],
            'packages': ['a_package'],
            'otherkey': 'some_other_key',
            'environment': {'test_env': '1', 'test_env2': '0'},
        }], disk_images)

    @mock.patch('yaml.safe_load', autospec=True)
    @mock.patch('os.path.isfile', autospec=True)
    def test_load_config_files_missing_image_name(self, mock_os_path_isfile,
                                                  mock_yaml_load):
        mock_yaml_load.return_value = {
            'disk_images': [{
                'arch': 'amd64',
                'imagename': 'overcloud',
                'type': 'qcow2',
                'elements': ['image_element']
            }, {
                'arch': 'amd64',
                'type': 'qcow2',
            }]
        }

        mock_os_path_isfile.return_value = True

        mock_open_context = mock.mock_open()
        mock_open_context().read.return_value = "YAML"

        with mock.patch('six.moves.builtins.open', mock_open_context):
            base_manager = BaseImageManager(['yamlfile'])
            self.assertRaises(ImageSpecificationException,
                              base_manager.load_config_files, 'disk_images')

    @mock.patch('yaml.safe_load', autospec=True)
    @mock.patch('os.path.isfile', autospec=True)
    def test_load_config_files_single_image(self, mock_os_path_isfile,
                                            mock_yaml_load):
        mock_yaml_load.side_effect = [{
            'disk_images': [
                {
                    'arch': 'amd64',
                    'imagename': 'overcloud',
                    'distro': 'some_awesome_distro',
                    'type': 'qcow2',
                    'elements': ['image_element']
                },
                {
                    'arch': 'amd64',
                    'imagename': 'not-overcloud',
                    'distro': 'some_other_distro',
                    'type': 'qcow2',
                    'elements': ['other_element']
                }
            ]}]

        mock_os_path_isfile.return_value = True

        mock_open_context = mock.mock_open()
        mock_open_context().read.return_value = "YAML"

        with mock.patch('six.moves.builtins.open', mock_open_context):
            base_manager = BaseImageManager(['yamlfile1'],
                                            images=['not-overcloud'])
            disk_images = base_manager.load_config_files('disk_images')

        self.assertEqual(1, mock_yaml_load.call_count)
        self.assertEqual([{
            'arch': 'amd64',
            'distro': 'some_other_distro',
            'imagename': 'not-overcloud',
            'type': 'qcow2',
            'elements': ['other_element'],
        }], disk_images)



# Copyright 2015 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import os
import re

from oslo_log import log
from oslo_utils import strutils

import tripleo_common.arch
from tripleo_common.image.base import BaseImageManager
from tripleo_common.image.exception import ImageSpecificationException
from tripleo_common.image.image_builder import ImageBuilder


class ImageBuildManager(BaseImageManager):
    """Manage the building of image files

       Manage the building of images from a config file specified in YAML
       syntax. Multiple config files can be specified. They will be merged
       """
    logger = log.getLogger(__name__ + '.ImageBuildManager')

    APPEND_ATTRIBUTES = BaseImageManager.APPEND_ATTRIBUTES + ['environment']

    def __init__(self, config_files, images=None, output_directory='.',
                 skip=False):
        super(ImageBuildManager, self).__init__(config_files, images)
        self.output_directory = re.sub('[/]$', '', output_directory)
        self.skip = skip

    def build(self):
        """Start the build process"""

        self.logger.info('Using config files: %s' % self.config_files)

        disk_images = self.load_config_files(self.DISK_IMAGES)

        for image in disk_images:
            arch = image.get('arch', tripleo_common.arch.dib_arch())
            image_type = image.get('type', 'qcow2')
            image_name = image.get('imagename')
            builder = image.get('builder', 'dib')
            skip_base = strutils.bool_from_string(
                image.get('skip_base', False))
            docker_target = image.get('docker_target')
            node_dist = image.get('distro')
            if node_dist is None:
                raise ImageSpecificationException('distro is required')
            self.logger.info('imagename: %s' % image_name)
            image_extension = image.get('imageext', image_type)
            image_path = os.path.join(self.output_directory, image_name)
            if self.skip:
                self.logger.info('looking for image at path: %s' % image_path)
                if os.path.exists('%s.%s' % (image_path, image_extension)):
                    self.logger.info('Image file exists for image name: %s' %
                                     image_name)
                    self.logger.info('Skipping image build')
                    continue
            elements = image.get('elements', [])
            options = image.get('options', [])
            packages = image.get('packages', [])
            environment = image.get('environment', {})

            extra_options = {
                'skip_base': skip_base,
                'docker_target': docker_target,
                'environment': environment
            }

            builder = ImageBuilder.get_builder(builder)
            builder.build_image(image_path, image_type, node_dist, arch,
                                elements, options, packages, extra_options)



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
import fixtures
import os
from unittest.mock import MagicMock

import yaml

from tripleo_common.tests import base
from tripleo_common.inventories import TripleoInventories


class _TestInventoriesBase(base.TestCase):
    def setUp(self):
        super(_TestInventoriesBase, self).setUp()
        self.read_inventory_data()

    def read_inventory_data(self):
        inventory_data = collections.OrderedDict()
        inventory_dir = os.path.join(
            os.path.dirname(__file__), 'inventory_data'
        )
        for datafile in (
                'cell1_dynamic.json',
                'cell1_static.yaml',
                'overcloud_dynamic.json',
                'overcloud_static.yaml',
                'merged_dynamic.json',
                'merged_static.yaml',
                'single_dynamic.json',
                'single_static.yaml',
                'undercloud_dynamic.json',
                'undercloud_static.yaml',
                'undercloud_dynamic_merged.json',
                'undercloud_static_merged.yaml',
                ):
            name = os.path.basename(datafile).split('.')[0]
            path = os.path.join(inventory_dir, datafile)
            with open(path, 'r') as data:
                inventory_data[name] = yaml.safe_load(data)
        self.inventory_data = inventory_data


class TestInventoriesStatic(_TestInventoriesBase):
    def setUp(self):
        super(TestInventoriesStatic, self).setUp()
        mock_inv_overcloud = MagicMock()
        mock_inv_cell1 = MagicMock()
        mock_inv_overcloud.list.return_value = self.inventory_data[
            'overcloud_static'
        ]
        mock_inv_cell1.list.return_value = self.inventory_data[
            'cell1_static'
        ]
        stack_to_inv_obj_map = {
            'overcloud': mock_inv_overcloud,
            'cell1': mock_inv_cell1
            }
        self.inventories = TripleoInventories(stack_to_inv_obj_map)

    def test_merge(self):
        actual = dict(self.inventories._merge(dynamic=False))
        expected = self.inventory_data['merged_static']
        self.assertEqual(expected, actual)

    def test_inventory_write_static(self):
        tmp_dir = self.useFixture(fixtures.TempDir()).path
        inv_path = os.path.join(tmp_dir, "inventory.yaml")
        self.inventories.write_static_inventory(inv_path)
        expected = self.inventory_data['merged_static']
        with open(inv_path, 'r') as f:
            loaded_inv = collections.OrderedDict(yaml.safe_load(f))
        self.assertEqual(expected, loaded_inv)


class TestInventoriesDynamic(_TestInventoriesBase):
    def setUp(self):
        super(TestInventoriesDynamic, self).setUp()
        mock_inv_overcloud = MagicMock()
        mock_inv_cell1 = MagicMock()
        mock_inv_overcloud.list.return_value = self.inventory_data[
            'overcloud_dynamic'
        ]
        mock_inv_cell1.list.return_value = self.inventory_data[
            'cell1_dynamic'
        ]
        stack_to_inv_obj_map = {
            'overcloud': mock_inv_overcloud,
            'cell1': mock_inv_cell1
            }
        self.inventories = TripleoInventories(stack_to_inv_obj_map)

    def test_merge(self):
        actual = dict(self.inventories._merge())
        expected = dict(self.inventory_data['merged_dynamic'])
        self.assertEqual(expected, actual)

    def test_list(self):
        actual = self.inventories.list()
        expected = self.inventory_data['merged_dynamic']
        self.assertEqual(expected, actual)


class TestInventorySingleStatic(_TestInventoriesBase):
    def setUp(self):
        super(TestInventorySingleStatic, self).setUp()
        mock_inv_overcloud = MagicMock()
        mock_inv_overcloud.list.return_value = self.inventory_data[
            'overcloud_static'
        ]
        stack_to_inv_obj_map = {
            'overcloud': mock_inv_overcloud
        }
        self.inventories = TripleoInventories(stack_to_inv_obj_map)

    def test_list(self):
        actual = self.inventories.list()
        expected = self.inventory_data['single_static']
        self.assertEqual(expected, actual)


class TestInventorySingleDynamic(_TestInventoriesBase):
    def setUp(self):
        super(TestInventorySingleDynamic, self).setUp()
        mock_inv_overcloud = MagicMock()
        mock_inv_overcloud.list.return_value = self.inventory_data[
            'overcloud_dynamic'
        ]
        stack_to_inv_obj_map = {
            'overcloud': mock_inv_overcloud
        }
        self.inventories = TripleoInventories(stack_to_inv_obj_map)

    def test_list(self):
        actual = self.inventories.list()
        expected = self.inventory_data['single_dynamic']
        self.assertEqual(expected, actual)


class TestInventoryUndercloudStatic(_TestInventoriesBase):
    def setUp(self):
        super(TestInventoryUndercloudStatic, self).setUp()
        mock_inv_undercloud = MagicMock()
        mock_inv_undercloud.list.return_value = self.inventory_data[
            'undercloud_static'
        ]
        stack_to_inv_obj_map = {
            'foobar': mock_inv_undercloud
        }
        self.inventories = TripleoInventories(stack_to_inv_obj_map)

    def test_list(self):
        actual = self.inventories.list(dynamic=False)
        expected = self.inventory_data['undercloud_static_merged']
        self.assertEqual(expected, actual)


class TestInventoryUndercloudDynamic(_TestInventoriesBase):
    def setUp(self):
        super(TestInventoryUndercloudDynamic, self).setUp()
        mock_inv_undercloud = MagicMock()
        mock_inv_undercloud.list.return_value = self.inventory_data[
            'undercloud_dynamic'
        ]
        stack_to_inv_obj_map = {
            'foobar': mock_inv_undercloud
        }
        self.inventories = TripleoInventories(stack_to_inv_obj_map)

    def test_list(self):
        actual = self.inventories.list()
        expected = self.inventory_data['undercloud_dynamic_merged']
        self.assertEqual(expected, actual)
